// Precomputed Atmospheric Scattering
// Copyright (C) 2008 Eric Bruneton
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. Neither the name of the copyright holders nor the names of its
//    contributors may be used to endorse or promote products derived from
//    this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
// THE POSSIBILITY OF SUCH DAMAGE.

#ifdef _FRAGMENT_

//-----------------------------------------------------------------------------

const int RES_R = 32;
const int RES_MU = 64; // 128
const int RES_MU_S = 32;
const int RES_NU = 8;

const float pi = 3.14159265359;

//-----------------------------------------------------------------------------
// Uniforms

uniform sampler2D irradianceSampler;    // precomputed skylight irradiance (E table)
uniform sampler2D transmittanceSampler; // precomputed transmittance (T table)
uniform sampler3D inscatterSampler;     // precomputed inscattered light (S table)

uniform vec4 AtmoParams1;   // (density, scattering bright, skylight bright, exposure)
uniform vec4 AtmoParams2;   // (MieG, MieFade, HR, HM);
uniform vec3 AtmoRayleigh;  // (betaR)
uniform vec3 AtmoMieExt;    // (betaMExt);
uniform vec4 Radiuses;      // (atmosphere bottom radius, atmosphere top radius, atmosphere height, surface radius)

float atmoH;
float atmoH2;
float planRadius2;
float mieG2;
float HorizonFixEps;

//-----------------------------------------------------------------------------

vec4 texture4D(float r, float mu, float muS, float nu)
{
    float rho2 = r * r - planRadius2;
    float rho = sqrt(rho2);
    float rmu = r * mu;
    float delta = rmu * rmu - rho2;
    vec4  cst = (rmu < 0.0 && delta > 0.0) ? vec4(1.0, 0.0, 0.0, 0.5 - 0.5 / float(RES_MU)) : vec4(-1.0, atmoH2, atmoH, 0.5 + 0.5 / float(RES_MU));
    float uR = 0.5 / float(RES_R) + rho / atmoH * (1.0 - 1.0 / float(RES_R));
    float uMu = cst.w + (rmu * cst.x + sqrt(delta + cst.y)) / (rho + cst.z) * (0.5 - 1.0 / float(RES_MU));
    // paper formula
    float uMuS = 0.5 / float(RES_MU_S) + max((1.0 - exp(-3.0 * muS - 0.6)) / (1.0 - exp(-3.6)), 0.0) * (1.0 - 1.0 / float(RES_MU_S));
    // better formula
    // float uMuS = 0.5 / float(RES_MU_S) + (atan(max(muS, -0.1975) * tan(1.26 * 1.1)) / 1.1 + (1.0 - 0.26)) * 0.5 * (1.0 - 1.0 / float(RES_MU_S));
    float lrp = (nu + 1.0) / 2.0 * (float(RES_NU) - 1.0);
    float uNu = floor(lrp);
    lrp -= uNu;
    return  mix(texture3D(inscatterSampler, vec3((uNu + uMuS) / float(RES_NU), uMu, uR)),
                texture3D(inscatterSampler, vec3((uNu + uMuS + 1.0) / float(RES_NU), uMu, uR)), lrp);
}

//-----------------------------------------------------------------------------
// Rayleigh phase function
float phaseFunctionR(float mu) 
{
    return 3.0 / (16.0 * pi) * (1.0 + mu*mu);
}

//-----------------------------------------------------------------------------
// Mie phase function
float phaseFunctionM(float mu) 
{
    return 3.0 / (8.0 * pi) * (1.0 - mieG2) * pow(1.0 + mieG2 - 2.0*AtmoParams2.r*mu, -1.5) * (1.0 + mu*mu) / (2.0 + mieG2);
}

//-----------------------------------------------------------------------------
// approximated single Mie scattering (cf. approximate Cm in paragraph
// "Angular precision")  rayMie.rgb=C*, rayMie.w=Cm,r
vec3 getMie(vec4 rayMie) 
{
    return rayMie.xyz * rayMie.w / max(rayMie.x, 1e-4) * (AtmoRayleigh.r / AtmoRayleigh.rgb);
}

//-----------------------------------------------------------------------------
// irradiance by sky light
// h - relative height (h = (r - planRadius) / atmoHeight)
vec3 irradiance(float h, float muS) 
{
    float uMuS = (muS + 0.2) / (1.0 + 0.2);
    return texture2D(irradianceSampler, vec2(uMuS, h)).rgb;
}

//-----------------------------------------------------------------------------
// transmittance(=transparency) of atmosphere for infinite ray (r,mu)
// (mu=cos(view zenith angle)), intersections with ground ignored
// h - relative height (h = (r - planRadius) / atmoHeight)
vec3 transmittance(float h, float mu) 
{
    //float uMu = atan((mu + 0.15) / (1.0 + 0.15) * tan(1.5)) / 1.5;
    float uMu = atan((mu + 0.18) / (1.0 + 0.18) * tan(1.5)) / 1.5;
    return texture2D(transmittanceSampler, vec2(uMu, sqrt(h))).rgb;
}

//-----------------------------------------------------------------------------
// transmittance(=transparency) of atmosphere for infinite ray (r,mu)
// (mu=cos(view zenith angle)), or zero if ray intersects ground
// h - relative height (h = (r - planRadius) / atmoHeight)
vec3 transmittanceWithShadow(float h, float mu, float mu0) 
{
    return (mu < mu0) ? vec3(0.0) : transmittance(h, mu);
}

//-----------------------------------------------------------------------------
// transmittance(=transparency) of atmosphere for infinite ray (r,mu)
// (mu=cos(view zenith angle)), or zero if ray intersects ground
// h - relative height (h = (r - planRadius) / atmoHeight)
vec3 transmittanceWithShadowDens(float h, float mu, float mu0) 
{
    return (mu < mu0) ? vec3(0.0) : mix(vec3(1.0), transmittance(h, mu), AtmoParams1.x);
}

//-----------------------------------------------------------------------------
// transmittance(=transparency) of atmosphere between x and x0
// assume segment x,x0 not intersecting ground
// r=||x||, mu=cos(zenith angle of [x,x0) ray at x)
vec3 transmittance(float h, float mu, float h0, float mu0)
{
    vec3 tr;
    if (mu > 0.0)
        tr = transmittance(h, mu) / transmittance(h0, mu0);
    else
        tr = transmittance(h0, -mu0) / transmittance(h, -mu);
    return mix(vec3(1.0), tr, AtmoParams1.x);
}

//-----------------------------------------------------------------------------
// optical depth for ray (r,mu) of length d, using analytic formula
// (mu=cos(view zenith angle)), intersections with ground ignored
// H = height scale of exponential density function
float opticalDepth(float H, float r, float mu, float d)
{
    float a = sqrt(0.5*r/H);
    vec2  a01 = a * vec2(mu, mu + d / r);
    vec2  a01s = sign(a01);
    vec2  a01sq = a01*a01;
    float x = a01s.y > a01s.x ? exp(a01sq.x) : 0.0;
    vec2  y = a01s / (2.3193*abs(a01) + sqrt(1.52*a01sq + 4.0)) * vec2(1.0, exp(-d/H*(d/(2.0*r)+mu)));
    return sqrt(6.2831*H*r) * exp((Radiuses.x-r)/H) * (x + dot(y, vec2(1.0, -1.0)));
}

//-----------------------------------------------------------------------------
// transmittance(=transparency) of atmosphere for ray (r,mu) of length d
// (mu=cos(view zenith angle)), intersections with ground ignored
// uses analytic formula instead of transmittance texture
vec3 transmittanceAnalytic(float r, float mu, float d)
{
    //mu = max(mu, -0.15);
    //mu = max(mu, 0.0);
    vec3 tr = exp(-AtmoRayleigh.rgb * opticalDepth(AtmoParams2.z, r, mu, d) - AtmoMieExt.rgb * opticalDepth(AtmoParams2.w, r, mu, d));
    return mix(vec3(1.0), tr, AtmoParams1.x);
}

//-----------------------------------------------------------------------------
//S[L]|x - T(x,xs)S[l]|xs = S[L]|x for spherical ground	
//inscattered light along ray x+tv, when sun in direction s (=S[L]|x)
vec3 inscatter(vec3 lightVec)
{
    float  nu = dot(eyeVec, lightVec);
    float  EyeMuS = dot(EyePosM, lightVec) / EyeR;
    float  phaseR = phaseFunctionR(nu);
    float  phaseM = phaseFunctionM(nu);
    vec4   inScatter = max(texture4D(EyeR, EyeMu, EyeMuS, nu), 0.0);

    // Avoids imprecision problems in Mie scattering when sun is below horizon
    inScatter.w *= smoothstep(0.0, AtmoParams2.y, EyeMuS);

    return inScatter.rgb * phaseR + getMie(inScatter) * phaseM;
}

//-----------------------------------------------------------------------------

vec3 inscatterGround(vec3 lightVec, vec3 Attenuation)
{
    float nu = dot(eyeVec, lightVec);
    float phaseR = phaseFunctionR(nu);
    float phaseM = phaseFunctionM(nu);
    float EyeMuS  = dot(EyePosM, lightVec) / EyeR;
    float FragMuS = dot(FragPos, lightVec) / FragR;

    vec4  eyeInScatter  = texture4D(EyeR,  EyeMu,  EyeMuS,  nu);
    vec4  fragInScatter = texture4D(FragR, FragMu, FragMuS, nu);
    vec4  inScatter = max(eyeInScatter - Attenuation.rgbr * fragInScatter, 0.0);

    // Avoids imprecision problems in Mie scattering when sun is below horizon
    inScatter.a *= smoothstep(0.0, AtmoParams2.y, EyeMuS);

    return inScatter.rgb * phaseR + getMie(inScatter) * phaseM;
}

//-----------------------------------------------------------------------------

vec3 inscatterGroundFix(vec3 lightVec, vec3 Attenuation)
{
    float nu = dot(eyeVec, lightVec);
    float phaseR = phaseFunctionR(nu);
    float phaseM = phaseFunctionM(nu);
    float EyeMuS  = dot(EyePosM, lightVec) / EyeR;
    float FragMuS = dot(FragPos, lightVec) / FragR;
    vec4  eyeInScatter, fragInScatter;

    // Avoids imprecision problems near horizon by interpolating between two points above and below horizon
    float a = ((EyeMu - HorizontMu) + HorizonFixEps) / (2.0 * HorizonFixEps);

    EyeMu = HorizontMu - HorizonFixEps;
    FragR = sqrt(EyeR * EyeR + eyeVecLength * eyeVecLength + 2.0 * EyeR * eyeVecLength * EyeMu);
    FragMu = (EyeR * EyeMu + eyeVecLength) / FragR;
    eyeInScatter  = texture4D(EyeR,  EyeMu,  EyeMuS,  nu);
    fragInScatter = texture4D(FragR, FragMu, FragMuS, nu);
    vec4 inScatterA = max(eyeInScatter - Attenuation.rgbr * fragInScatter, 0.0);

    EyeMu = HorizontMu + HorizonFixEps;
    FragR = sqrt(EyeR * EyeR + eyeVecLength * eyeVecLength + 2.0 * EyeR * eyeVecLength * EyeMu);
    FragMu = (EyeR * EyeMu + eyeVecLength) / FragR;
    eyeInScatter  = texture4D(EyeR,  EyeMu,  EyeMuS,  nu);
    fragInScatter = texture4D(FragR, FragMu, FragMuS, nu);
    vec4 inScatterB = max(eyeInScatter - Attenuation.rgbr * fragInScatter, 0.0);

    vec4 inScatter = mix(inScatterA, inScatterB, a);

    // Avoids imprecision problems in Mie scattering when sun is below horizon
    inScatter.w *= smoothstep(0.0, AtmoParams2.y, EyeMuS);

    return inScatter.rgb * phaseR + getMie(inScatter) * phaseM;
}

//-----------------------------------------------------------------------------

#endif
