// Real-time DXT1 & YCoCg DXT5 compression (GLSL)
// Copyright (c) NVIDIA Corporation.
// Written by: Ignacio Castano <icastano@nvidia.com>
// 
// Thanks to JMP van Waveren, Simon Green, Eric Werness, Simon Brown
// 
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use,
// copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.

#extension GL_EXT_gpu_shader4 : enable

#ifdef _FRAGMENT_

const float offset = 128.0 / 255.0;

//-----------------------------------------------------------------------------

// Use dot product to minimize RMS instead absolute distance like in the CPU compressor.
float colorDistance(vec3 c0, vec3 c1)
{
    return dot(c0-c1, c0-c1);
}

float colorDistance(vec2 c0, vec2 c1)
{
    return dot(c0-c1, c0-c1);
}

vec3 toYCoCg(vec3 c)
{
    float Y  = (c.r + 2 * c.g + c.b) * 0.25;
    float Co = (( 2 * c.r - 2 * c.b      ) * 0.25 + offset);
    float Cg = ((    -c.r + 2 * c.g - c.b) * 0.25 + offset);
    return vec3(Y, Co, Cg);
}

//-----------------------------------------------------------------------------

void    ExtractColorBlockXY(out vec2 col[16], sampler2D image, vec2 texcoord, vec2 imageSize)
{
#if 0
    vec2 texelSize = (1.0f / imageSize);
    texcoord -= texelSize * 2;
    for (int i=0; i<4; i++)
        for (int j=0; j<4; j++)
            col[i*4+j] = texture2D(image, texcoord + vec2(j, i) * texelSize).rg;
#else
    // use TXF instruction (integer coordinates with offset)
    // note offsets must be constant
    //ivec4 base = ivec4(wpos*4-2, 0, 0);
    ivec2 base = ivec2(texcoord * imageSize - 1.5);
    col[0]  = texelFetch2DOffset(image, base, 0, ivec2(0, 0)).rg;
    col[1]  = texelFetch2DOffset(image, base, 0, ivec2(1, 0)).rg;
    col[2]  = texelFetch2DOffset(image, base, 0, ivec2(2, 0)).rg;
    col[3]  = texelFetch2DOffset(image, base, 0, ivec2(3, 0)).rg;
    col[4]  = texelFetch2DOffset(image, base, 0, ivec2(0, 1)).rg;
    col[5]  = texelFetch2DOffset(image, base, 0, ivec2(1, 1)).rg;
    col[6]  = texelFetch2DOffset(image, base, 0, ivec2(2, 1)).rg;
    col[7]  = texelFetch2DOffset(image, base, 0, ivec2(3, 1)).rg;
    col[8]  = texelFetch2DOffset(image, base, 0, ivec2(0, 2)).rg;
    col[9]  = texelFetch2DOffset(image, base, 0, ivec2(1, 2)).rg;
    col[10] = texelFetch2DOffset(image, base, 0, ivec2(2, 2)).rg;
    col[11] = texelFetch2DOffset(image, base, 0, ivec2(3, 2)).rg;
    col[12] = texelFetch2DOffset(image, base, 0, ivec2(0, 3)).rg;
    col[13] = texelFetch2DOffset(image, base, 0, ivec2(1, 3)).rg;
    col[14] = texelFetch2DOffset(image, base, 0, ivec2(2, 3)).rg;
    col[15] = texelFetch2DOffset(image, base, 0, ivec2(3, 3)).rg;
#endif
}

//-----------------------------------------------------------------------------

void ExtractColorBlockRGB(out vec3 col[16], sampler2D image, vec2 texcoord, vec2 imageSize)
{
#if 0
    vec2 texelSize = (1.0f / imageSize);
    texcoord -= texelSize * 2;
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            col[i*4+j] = texture2D(image, texcoord + vec2(j, i) * texelSize).rgb;
        }
    }
#else
    // use TXF instruction (integer coordinates with offset)
    // note offsets must be constant
    //ivec4 base = ivec4(wpos*4-2, 0, 0);
    ivec2 base = ivec2(texcoord * imageSize - 1.5);
    col[0]  = texelFetch2DOffset(image, base, 0, ivec2(0, 0)).rgb;
    col[1]  = texelFetch2DOffset(image, base, 0, ivec2(1, 0)).rgb;
    col[2]  = texelFetch2DOffset(image, base, 0, ivec2(2, 0)).rgb;
    col[3]  = texelFetch2DOffset(image, base, 0, ivec2(3, 0)).rgb;
    col[4]  = texelFetch2DOffset(image, base, 0, ivec2(0, 1)).rgb;
    col[5]  = texelFetch2DOffset(image, base, 0, ivec2(1, 1)).rgb;
    col[6]  = texelFetch2DOffset(image, base, 0, ivec2(2, 1)).rgb;
    col[7]  = texelFetch2DOffset(image, base, 0, ivec2(3, 1)).rgb;
    col[8]  = texelFetch2DOffset(image, base, 0, ivec2(0, 2)).rgb;
    col[9]  = texelFetch2DOffset(image, base, 0, ivec2(1, 2)).rgb;
    col[10] = texelFetch2DOffset(image, base, 0, ivec2(2, 2)).rgb;
    col[11] = texelFetch2DOffset(image, base, 0, ivec2(3, 2)).rgb;
    col[12] = texelFetch2DOffset(image, base, 0, ivec2(0, 3)).rgb;
    col[13] = texelFetch2DOffset(image, base, 0, ivec2(1, 3)).rgb;
    col[14] = texelFetch2DOffset(image, base, 0, ivec2(2, 3)).rgb;
    col[15] = texelFetch2DOffset(image, base, 0, ivec2(3, 3)).rgb;
#endif
}

//-----------------------------------------------------------------------------

void ExtractColorBlockYCoCg(out vec3 col[16], sampler2D image, vec2 texcoord, vec2 imageSize)
{
#if 0
    vec2 texelSize = (1.0f / imageSize);
    texcoord -= texelSize * 2;
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            col[i*4+j] = toYCoCg(texture2D(image, texcoord + vec2(j, i) * texelSize).rgb);
        }
    }
#else
    // use TXF instruction (integer coordinates with offset)
    // note offsets must be constant
    //ivec4 base = ivec4(wpos*4-2, 0, 0);
    ivec2 base = ivec2(texcoord * imageSize - 1.5);
    col[0]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(0, 0)).rgb);
    col[1]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(1, 0)).rgb);
    col[2]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(2, 0)).rgb);
    col[3]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(3, 0)).rgb);
    col[4]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(0, 1)).rgb);
    col[5]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(1, 1)).rgb);
    col[6]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(2, 1)).rgb);
    col[7]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(3, 1)).rgb);
    col[8]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(0, 2)).rgb);
    col[9]  = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(1, 2)).rgb);
    col[10] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(2, 2)).rgb);
    col[11] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(3, 2)).rgb);
    col[12] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(0, 3)).rgb);
    col[13] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(1, 3)).rgb);
    col[14] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(2, 3)).rgb);
    col[15] = toYCoCg(texelFetch2DOffset(image, base, 0, ivec2(3, 3)).rgb);
#endif
}

//-----------------------------------------------------------------------------

// find minimum and maximum colors based on bounding box in color space
void    FindMinMaxColorsBox(vec2 block[16], out vec2 mincol, out vec2 maxcol)
{
    mincol = block[0];
    maxcol = block[0];
    
    for (int i=1; i<16; i++)
    {
        mincol = min(mincol, block[i]);
        maxcol = max(maxcol, block[i]);
    }
}

//-----------------------------------------------------------------------------

// find minimum and maximum colors based on bounding box in color space
void    FindMinMaxColorsBox(vec3 block[16], out vec3 mincol, out vec3 maxcol)
{
    mincol = block[0];
    maxcol = block[0];
    
    for (int i = 1; i < 16; i++)
    {
        mincol = min(mincol, block[i]);
        maxcol = max(maxcol, block[i]);
    }
}

//-----------------------------------------------------------------------------

void    InsetNormalsBBoxDXT5(in out vec2 mincol, in out vec2 maxcol)
{
    vec2 inset;
    inset.x = (maxcol.x - mincol.x) / 32.0 - (16.0 / 255.0) / 32.0; // ALPHA scale-bias.
    inset.y = (maxcol.y - mincol.y) / 16.0 - (8.0 / 255.0) / 16;    // GREEN scale-bias.
    mincol = clamp(mincol + inset, 0.0, 1.0);
    maxcol = clamp(maxcol - inset, 0.0, 1.0);
}

//-----------------------------------------------------------------------------

void    InsetNormalsBBoxLATC(in out vec2 mincol, in out vec2 maxcol)
{
    vec2 inset = (maxcol - mincol) / 32.0 - (16.0 / 255.0) / 32.0;  // ALPHA scale-bias.
    mincol = clamp(mincol + inset, 0.0, 1.0);
    maxcol = clamp(maxcol - inset, 0.0, 1.0);
}

//-----------------------------------------------------------------------------

void InsetBBox(in out vec3 mincol, in out vec3 maxcol)
{
    vec3 inset = (maxcol - mincol) / 16.0 - (8.0 / 255.0) / 16;
    mincol = clamp(mincol + inset, 0.0, 1.0);
    maxcol = clamp(maxcol - inset, 0.0, 1.0);
}

//-----------------------------------------------------------------------------

void InsetYBBox(in out float mincol, in out float maxcol)
{
    float inset = (maxcol - mincol) / 32.0 - (16.0 / 255.0) / 32.0;
    mincol = clamp(mincol + inset, 0.0, 1.0);
    maxcol = clamp(maxcol - inset, 0.0, 1.0);
}

//-----------------------------------------------------------------------------

void InsetCoCgBBox(in out vec2 mincol, in out vec2 maxcol)
{
    vec2 inset = (maxcol - mincol) / 16.0 - (8.0 / 255.0) / 16;
    mincol = clamp(mincol + inset, 0.0, 1.0);
    maxcol = clamp(maxcol - inset, 0.0, 1.0);
}

//-----------------------------------------------------------------------------

uint    EmitGreenEndPoints(in out float ming, in out float maxg)
{
    uint c0 = uint(round(ming * 63.0));
    uint c1 = uint(round(maxg * 63.0));

    ming = float((c0 << uint(2)) | (c0 >> uint(4))) * (1.0 / 255.0);
    maxg = float((c1 << uint(2)) | (c1 >> uint(4))) * (1.0 / 255.0);

    return (c0 << uint(21)) | (c1 << uint(5));
}

//-----------------------------------------------------------------------------

#if 1

uint    EmitGreenIndices(vec2 block[16], float minGreen, float maxGreen)
{
    const int GREEN_RANGE = 3;

    float bias = maxGreen + (maxGreen - minGreen) / (2.0 * GREEN_RANGE);
    float scale = 1.0f / (maxGreen - minGreen);

    // Compute indices
    uint indices = uint(0);
    for (int i=0; i<16; i++)
    {
        uint index = uint(clamp((bias - block[i].y) * scale, 0.0, 1.0) * GREEN_RANGE);
        indices |= index << uint(i * 2);
    }

    uint i0 = indices & uint(0x55555555);
    uint i1 = (indices & uint(0xAAAAAAAA)) >> uint(1);
    indices = ((i0 ^ i1) << uint(1)) | i1;

    // Output indices
    return indices;
}

#else

uint EmitGreenIndices(vec2 block[16], float minGreen, float maxGreen)
{
    const int GREEN_RANGE = 3;

    float mid = (maxGreen - minGreen) / (2 * GREEN_RANGE);

    float yb1 = minGreen + mid;
    float yb2 = (2 * maxGreen + 1 * minGreen) / GREEN_RANGE + mid;
    float yb3 = (1 * maxGreen + 2 * minGreen) / GREEN_RANGE + mid;

    // Compute indices
    uint indices = 0;
    for (int i=0; i<16; i++)
    {
        float y = block[i].y;

        uint index = (y <= yb1);
        index += (y <= yb2);
        index += (y <= yb3);

        indices |= index << (i * 2);
    }

    uint i0 = (indices & 0x55555555);
    uint i1 = (indices & 0xAAAAAAAA) >> 1;
    indices = ((i0 ^ i1) << 1) | i1;

    // Output indices
    return indices;
}

#endif

//-----------------------------------------------------------------------------

uint    EmitAlphaEndPoints(float mincol, float maxcol)
{
    uint c0 = uint(round(mincol * 255));
    uint c1 = uint(round(maxcol * 255));

    return (c0 << uint(8)) | c1;
}

//-----------------------------------------------------------------------------

uvec2   EmitAlphaIndices(vec2 block[16], float minAlpha, float maxAlpha)
{
    const int ALPHA_RANGE = 7;

    float bias = maxAlpha + (maxAlpha - minAlpha) / (2.0 * ALPHA_RANGE);
    float scale = 1.0f / (maxAlpha - minAlpha);

    uvec2 indices = uvec2(0, 0);

    for (int i=0; i<6; i++)
    {
        uint index = uint(clamp((bias - block[i].x) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.x |= index << uint(3 * i);
    }

    for (int i=6; i<16; i++)
    {
        uint index = uint(clamp((bias - block[i].x) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.y |= index << uint(3 * i - 18);
    }

    uvec2 i0 = (indices >> uint(0)) & uint(0x09249249);
    uvec2 i1 = (indices >> uint(1)) & uint(0x09249249);
    uvec2 i2 = (indices >> uint(2)) & uint(0x09249249);

    i2 ^= i0 & i1;
    i1 ^= i0;
    i0 ^= (i1 | i2);

    indices.x = (i2.x << uint(2)) | (i1.x << uint(1)) | i0.x;
    indices.y = (((i2.y << uint(2)) | (i1.y << uint(1)) | i0.y) << uint(2)) | (indices.x >> uint(16));
    indices.x <<= uint(16);

    return indices;
}

//-----------------------------------------------------------------------------

uvec2   EmitLuminanceIndices(vec2 block[16], float minAlpha, float maxAlpha)
{
    const int ALPHA_RANGE = 7;

    float bias = maxAlpha + (maxAlpha - minAlpha) / (2.0 * ALPHA_RANGE);
    float scale = 1.0f / (maxAlpha - minAlpha);

    uvec2 indices = uvec2(0, 0);

    for (int i=0; i<6; i++)
    {
        uint index = uint(clamp((bias - block[i].y) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.x |= index << uint(3 * i);
    }

    for (int i=6; i<16; i++)
    {
        uint index = uint(clamp((bias - block[i].y) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.y |= index << uint(3 * i - 18);
    }

    uvec2 i0 = (indices >> uint(0)) & uint(0x09249249);
    uvec2 i1 = (indices >> uint(1)) & uint(0x09249249);
    uvec2 i2 = (indices >> uint(2)) & uint(0x09249249);

    i2 ^= i0 & i1;
    i1 ^= i0;
    i0 ^= (i1 | i2);

    indices.x = (i2.x << uint(2)) | (i1.x << uint(1)) | i0.x;
    indices.y = (((i2.y << uint(2)) | (i1.y << uint(1)) | i0.y) << uint(2)) | (indices.x >> uint(16));
    indices.x <<= uint(16);

    return indices;
}

//-----------------------------------------------------------------------------

void SelectDiagonal(vec3 block[16], in out vec3 mincol, in out vec3 maxcol)
{
    vec3 center = (mincol + maxcol) * 0.5;

    vec2 cov = vec2(0);
    for (int i = 0; i < 16; i++)
    {
        vec3 t = block[i] - center;
        cov.x += t.x * t.z;
        cov.y += t.y * t.z;
    }

    if (cov.x < 0) {
        float temp = maxcol.x;
        maxcol.x = mincol.x;
        mincol.x = temp;
    }
    if (cov.y < 0) {
        float temp = maxcol.y;
        maxcol.y = mincol.y;
        mincol.y = temp;
    }
}

//-----------------------------------------------------------------------------

vec3 RoundAndExpand(vec3 v, out uint w)
{
    ivec3 c = ivec3(round(v * vec3(31, 63, 31)));
    w = uint((c.r << 11) | (c.g << 5) | c.b);

    c.rb = (c.rb << 3) | (c.rb >> 2);
    c.g = (c.g << 2) | (c.g >> 4);

    return vec3(c) * (1.0 / 255.0);
}

//-----------------------------------------------------------------------------

uint    EmitEndPointsDXT1(in out vec3 mincol, in out vec3 maxcol)
{
    uvec2 output;
    maxcol = RoundAndExpand(maxcol, output.x);
    mincol = RoundAndExpand(mincol, output.y);

    // We have to do this in case we select an alternate diagonal.
    if (output.x < output.y)
    {
        vec3 tmp = mincol;
        mincol = maxcol;
        maxcol = tmp;
        return output.y | (output.x << uint(16));
    }

    return output.x | (output.y << uint(16));
}

#if 1

uint    EmitIndicesDXT1(vec3 block[16], vec3 mincol, vec3 maxcol)
{
    const float RGB_RANGE = 3;

    vec3 dir = (maxcol - mincol);
    vec3 origin = maxcol + dir / (2.0 * RGB_RANGE);
    dir /= dot(dir, dir);

    // Compute indices
    uint indices = uint(0);
    for (int i=0; i<16; i++)
    {
        uint index = uint(clamp(dot(origin - block[i], dir), 0.0, 1.0) * RGB_RANGE);
        indices |= index << uint(i * 2);
    }

    uint i0 = (indices & uint(0x55555555));
    uint i1 = (indices & uint(0xAAAAAAAA)) >> uint(1);
    indices = ((i0 ^ i1) << uint(1)) | i1;

    // Output indices
    return indices;
}

#else

uint    EmitIndicesDXT1(vec3 col[16], vec3 mincol, vec3 maxcol)
{
    // Compute palette
    vec3 c[4];
    c[0] = maxcol;
    c[1] = mincol;
    c[2] = lerp(c[0], c[1], 1.0/3.0);
    c[3] = lerp(c[0], c[1], 2.0/3.0);

    // Compute indices
    uint indices = 0;
    for (int i = 0; i < 16; i++) {

        // find index of closest color
        vec4 dist;
        dist.x = colorDistance(col[i], c[0]);
        dist.y = colorDistance(col[i], c[1]);
        dist.z = colorDistance(col[i], c[2]);
        dist.w = colorDistance(col[i], c[3]);
        
        uint4 b = dist.xyxy > dist.wzzw;
        uint b4 = dist.z > dist.w;
        
        uint index = (b.x & b4) | (((b.y & b.z) | (b.x & b.w)) << 1);
        indices |= index << (i*2);
    }

    // Output indices
    return indices;
}

#endif

//-----------------------------------------------------------------------------

int ScaleYCoCg(vec2 minColor, vec2 maxColor)
{
    vec2 m0 = abs(minColor - offset);
    vec2 m1 = abs(maxColor - offset);

    float m = max(max(m0.x, m0.y), max(m1.x, m1.y));

    const float s0 = 64.0 / 255.0;
    const float s1 = 32.0 / 255.0;

    int scale = 1;
    if (m < s0) scale = 2;
    if (m < s1) scale = 4;

    return scale;
}

//-----------------------------------------------------------------------------

void SelectYCoCgDiagonal(const vec3 block[16], in out vec2 minColor, in out vec2 maxColor)
{
    vec2 mid = (maxColor + minColor) * 0.5;

    float cov = 0;
    for (int i = 0; i < 16; i++)
    {
        vec2 t = block[i].yz - mid;
        cov += t.x * t.y;
    }
    if (cov < 0) {
        float tmp = maxColor.y;
        maxColor.y = minColor.y;
        minColor.y = tmp;
    }
}

//-----------------------------------------------------------------------------

uint EmitEndPointsYCoCgDXT5(in out vec2 mincol, in out vec2 maxcol, int scale)
{
    maxcol = (maxcol - offset) * scale + offset;
    mincol = (mincol - offset) * scale + offset;

    InsetCoCgBBox(mincol, maxcol);

    maxcol = round(maxcol * vec2(31, 63));
    mincol = round(mincol * vec2(31, 63));

    ivec2 imaxcol = ivec2(maxcol);
    ivec2 imincol = ivec2(mincol);

    uvec2 output;
    output.x = uint((imaxcol.r << 11) | (imaxcol.g << 5) | (scale - 1));
    output.y = uint((imincol.r << 11) | (imincol.g << 5) | (scale - 1));

    imaxcol.r = (imaxcol.r << 3) | (imaxcol.r >> 2);
    imaxcol.g = (imaxcol.g << 2) | (imaxcol.g >> 4);
    imincol.r = (imincol.r << 3) | (imincol.r >> 2);
    imincol.g = (imincol.g << 2) | (imincol.g >> 4);

    maxcol = vec2(imaxcol) * (1.0 / 255.0);
    mincol = vec2(imincol) * (1.0 / 255.0);

    // Undo rescale.
    maxcol = (maxcol - offset) / scale + offset;
    mincol = (mincol - offset) / scale + offset;

    return output.x | (output.y << uint(16));
}

//-----------------------------------------------------------------------------

#if 1

uint EmitIndicesYCoCgDXT5(vec3 block[16], vec2 mincol, vec2 maxcol)
{
    const float COCG_RANGE = 3;

    vec2 dir = (maxcol - mincol);
    vec2 origin = maxcol + dir / (2.0 * COCG_RANGE);
    dir /= dot(dir, dir);

    // Compute indices
    uint indices = uint(0);
    for (int i = 0; i < 16; i++)
    {
        uint index = uint(clamp(dot(origin - block[i].yz, dir), 0.0, 1.0) * COCG_RANGE);
        indices |= index << uint(i * 2);
    }

    uint i0 = (indices & uint(0x55555555));
    uint i1 = (indices & uint(0xAAAAAAAA)) >> uint(1);
    indices = ((i0 ^ i1) << uint(1)) | i1;

    // Output indices
    return indices;
}

#else

uint EmitIndicesYCoCgDXT5(vec3 block[16], vec2 mincol, vec2 maxcol)
{
    // Compute palette
    vec2 c[4];
    c[0] = maxcol;
    c[1] = mincol;
    c[2] = lerp(c[0], c[1], 1.0/3.0);
    c[3] = lerp(c[0], c[1], 2.0/3.0);

    // Compute indices
    uint indices = 0;
    for (int i = 0; i < 16; i++)
    {
        // find index of closest color
        vec4 dist;
        dist.x = colorDistance(block[i].yz, c[0]);
        dist.y = colorDistance(block[i].yz, c[1]);
        dist.z = colorDistance(block[i].yz, c[2]);
        dist.w = colorDistance(block[i].yz, c[3]);

        uint4 b = dist.xyxy > dist.wzzw;
        uint b4 = dist.z > dist.w;

        uint index = (b.x & b4) | (((b.y & b.z) | (b.x & b.w)) << 1);
        indices |= index << (i*2);
    }

    // Output indices
    return indices;
}

#endif

//-----------------------------------------------------------------------------

uint EmitAlphaEndPointsYCoCgDXT5(float mincol, float maxcol)
{
    uint c0 = uint(round(mincol * 255));
    uint c1 = uint(round(maxcol * 255));
    return (c0 << uint(8)) | c1;
}

//-----------------------------------------------------------------------------

#if 1

// Optimized index selection.
uvec2 EmitAlphaIndicesYCoCgDXT5(vec3 block[16], float minAlpha, float maxAlpha)
{
    const int ALPHA_RANGE = 7;

    float bias = maxAlpha + (maxAlpha - minAlpha) / (2.0 * ALPHA_RANGE);
    float scale = 1.0f / (maxAlpha - minAlpha);

    uvec2 indices = uvec2(0, 0);

    for (int i=0; i<6; i++)
    {
        uint index = uint(clamp((bias - block[i].x) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.x |= index << uint(3 * i);
    }

    for (int i=6; i<16; i++)
    {
        uint index = uint(clamp((bias - block[i].x) * scale, 0.0, 1.0) * ALPHA_RANGE);
        indices.y |= index << uint(3 * i - 18);
    }

    uvec2 i0 = (indices >> uint(0)) & uint(0x09249249);
    uvec2 i1 = (indices >> uint(1)) & uint(0x09249249);
    uvec2 i2 = (indices >> uint(2)) & uint(0x09249249);

    i2 ^= i0 & i1;
    i1 ^= i0;
    i0 ^= (i1 | i2);

    indices.x = (i2.x << uint(2)) | (i1.x << uint(1)) | i0.x;
    indices.y = (((i2.y << uint(2)) | (i1.y << uint(1)) | i0.y) << uint(2)) | (indices.x >> uint(16));
    indices.x <<= uint(16);

    return indices;
}

#else

// Version shown in the YCoCg-DXT article.
uvec2 EmitAlphaIndicesYCoCgDXT5(vec3 block[16], float minAlpha, float maxAlpha)
{
    const int ALPHA_RANGE = 7;

    float mid = (maxAlpha - minAlpha) / (2.0 * ALPHA_RANGE);

    float ab1 = minAlpha + mid;
    float ab2 = (6 * maxAlpha + 1 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;
    float ab3 = (5 * maxAlpha + 2 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;
    float ab4 = (4 * maxAlpha + 3 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;
    float ab5 = (3 * maxAlpha + 4 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;
    float ab6 = (2 * maxAlpha + 5 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;
    float ab7 = (1 * maxAlpha + 6 * minAlpha) * (1.0 / ALPHA_RANGE) + mid;

    uvec2 indices = 0;

    uint index;
    for (int i = 0; i < 6; i++)
    {
        float a = block[i].x;
        index = 1;
        index += (a <= ab1);
        index += (a <= ab2);
        index += (a <= ab3);
        index += (a <= ab4);
        index += (a <= ab5);
        index += (a <= ab6);
        index += (a <= ab7);
        index &= 7;
        index ^= (2 > index);
        indices.x |= index << (3 * i + 16);
    }

    indices.y = index >> 1;

    for (int i = 6; i < 16; i++)
    {
        float a = block[i].x;
        index = 1;
        index += (a <= ab1);
        index += (a <= ab2);
        index += (a <= ab3);
        index += (a <= ab4);
        index += (a <= ab5);
        index += (a <= ab6);
        index += (a <= ab7);
        index &= 7;
        index ^= (2 > index);
        indices.y |= index << (3 * i - 16);
    }

    return indices;
}

#endif

//-----------------------------------------------------------------------------

#endif
